/*  GSsoft
 *  Copyright (C) 2002-2004  GSsoft Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdlib.h>
#include <string.h>

#include "Draw.h"
#include "Mem.h"
#include "Page.h"
#include "Texts.h"

#ifdef __WIN32__
#pragma warning(disable:4244)
#endif

typedef struct {
	u32 *data;
	tex0Info tex0;
} Cache;

Cache *Texts;
Cache *Texture;
int TextureCount;

int  CacheInit() {
	Texts = (Cache*)malloc(conf.cachesize * sizeof(Cache));
	memset(Texts, 0, conf.cachesize * sizeof(Cache));
	TextureCount = 0;

	return 0;
}

void CacheLoadTexture(int i) {
	int u, v;

	Texts[i].tex0 = *tex0;
	Texts[i].data = malloc(Texts[i].tex0.th*Texts[i].tex0.tw*4);
	for (v=0; v<Texts[i].tex0.th; v++) {
		for (u=0; u<Texts[i].tex0.tw; u++) {
			Texts[i].data[v*Texts[i].tex0.tw+u] = GetTexturePixel32(u, v);
		}
	}
}

void CacheFreeTexture(int i) {
	free(Texts[i].data); Texts[i].data = NULL;
}

int  CacheSetTexture() {
	int i;

	if (tex0->tbp0 == gs._gsfb[0].fbp ||
		tex0->tbp0 == gs._gsfb[1].fbp) {
		return -1;
	}

	for (i=0; i<conf.cachesize; i++) {
		if (memcmp(&Texts[i].tex0, tex0, sizeof(tex0Info)) == 0) {
			Texture = &Texts[i];
			return 0;
		}
		if (Texts[i].data == NULL) break;
	}
	if (i == conf.cachesize) {
		i = TextureCount++;
		if (TextureCount >= conf.cachesize) TextureCount = 0;
		CacheFreeTexture(i);
	}

#ifdef GS_LOG
	GS_LOG("CacheSetTexture: %dx%d: %d\n", tex0->tw, tex0->th, i);
#endif
	printf("CacheSetTexture: %x, %dx%d: %d\n", tex0->tbp0, tex0->tw, tex0->th, i);
	CacheLoadTexture(i);
	Texture = &Texts[i];
	return 0;
}

u32  CacheGetTexturePixel32(int u, int v) {
	if (u > Texture->tex0.tw) return 0;
	if (v > Texture->tex0.th) return 0;
	return Texture->data[v*Texture->tex0.tw+u];
}

void CacheShutdown() {
	int i;

	for (i=0; i<conf.cachesize; i++) {
		if (Texts[i].data) CacheFreeTexture(i);
	}
	free(Texts);
}


