/*  GSsoft
 *  Copyright (C) 2002-2004  GSsoft Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include "Draw.h"
#include "Mem.h"
#include "Page.h"

#ifdef __WIN32__
#pragma warning(disable:4244)
#endif

////////////////////
// 15bit

void _ScrBlit32to15S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u32 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u16*)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel32(xpos>>16, ypos>>16, bp, dispfb->fbw);
			*pscr++= ((c & 0x0000f8) <<  7) | /* 0x7c00 */
					 ((c & 0x00f800) >>  6) | /* 0x03e0 */
					 ((c & 0xf80000) >> 19);  /* 0x001f */
		}
	}
}

void _ScrBlit16to15S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u16 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u16*)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel16(xpos>>16, ypos>>16, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 10) | /* 0x7c00 */
					 ((c & 0x03e0)      ) | /* 0x03e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void _ScrBlit16Sto15S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u16 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u16*)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel16S(xpos>>16, ypos>>16, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 10) | /* 0x7c00 */
					 ((c & 0x03e0)      ) | /* 0x03e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void ScrBlit15S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	switch (dispfb->psm) {
		case 0x2: _ScrBlit16to15S(scr, dw, dh, sp, gsScr, bp); break;
		case 0xA: _ScrBlit16Sto15S(scr, dw, dh, sp, gsScr, bp); break;
		default:  _ScrBlit32to15S(scr, dw, dh, sp, gsScr, bp); break;
	}
}

////////////////////
// 16bit

void _ScrBlit32to16S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u32 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u16*)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel32(xpos>>16, ypos>>16, bp, dispfb->fbw);
			*pscr++= ((c & 0x0000f8) <<  8) | /* 0xf800 */
					 ((c & 0x00fc00) >>  5) | /* 0x07e0 */
					 ((c & 0xf80000) >> 19);  /* 0x001f */
		}
	}
}

void _ScrBlit16to16S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u16 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u16*)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel16(xpos>>16, ypos>>16, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 11) | /* 0xf800 */
					 ((c & 0x03e0) <<  1) | /* 0x07e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void _ScrBlit16Sto16S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u16 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u16*)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel16(xpos>>16, ypos>>16, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 11) | /* 0xf800 */
					 ((c & 0x03e0) <<  1) | /* 0x07e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void ScrBlit16S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	switch (dispfb->psm) {
		case 0x2: _ScrBlit16to16S(scr, dw, dh, sp, gsScr, bp); break;
		case 0xA: _ScrBlit16Sto16S(scr, dw, dh, sp, gsScr, bp); break;
		default:  _ScrBlit32to16S(scr, dw, dh, sp, gsScr, bp); break;
	}
}

////////////////////
// 32bit

void _ScrBlit32to32S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u8 *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u32 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u8 *)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel32(xpos>>16, ypos>>16, bp, dispfb->fbw);
			pscr[0] = (c >> 16) & 0xff;
			pscr[1] = (c >>  8) & 0xff;
			pscr[2] = c & 0xff;
			pscr+=4;
		}
	}
}

void _ScrBlit16to32S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u8  *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u16 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u8 *)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel16(xpos>>16, ypos>>16, bp, dispfb->fbw);
			pscr[0] = (c >> 7) & 0xf8;
			pscr[1] = (c >> 2) & 0xf8;
			pscr[2] = (c << 3) & 0xf8;
			pscr+=4;
		}
	}
}

void _ScrBlit16Sto32S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u8  *pscr;
	int x, y;
	int xpos, xinc;
	int ypos, yinc;
	u16 c;

	xinc = (gsScr->w << 16) / dw;
	yinc = (gsScr->h << 16) / dh;

	for (y=0, ypos=gsScr->y<<16; y<dh; y++, ypos+=yinc) {
		pscr = (u8 *)(scr + y * sp);
		for(x=0, xpos=gsScr->x<<16; x<dw; x++, xpos+= xinc) {
			c = readPixel16S(xpos>>16, ypos>>16, bp, dispfb->fbw);
			pscr[0] = (c >> 7) & 0xf8;
			pscr[1] = (c >> 2) & 0xf8;
			pscr[2] = (c << 3) & 0xf8;
			pscr+=4;
		}
	}
}

void ScrBlit32S(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	switch (dispfb->psm) {
		case 0x2: _ScrBlit16to32S(scr, dw, dh, sp, gsScr, bp); break;
		case 0xA: _ScrBlit16Sto32S(scr, dw, dh, sp, gsScr, bp); break;
		default:  _ScrBlit32to32S(scr, dw, dh, sp, gsScr, bp); break;
	}
}


////////////////////
// 15bit

#define _ScrBlitSET() \
	if (gsScr->w < dw) { \
		sx = (dw - gsScr->w) / 2; \
		w = gsScr->w; \
	} else { \
		sx = 0; \
		w = dw; \
	} \
 \
	if (gsScr->h < dh) { \
		sy = (dh - gsScr->h) / 2; \
		h = gsScr->h; \
	} else { \
		sy = 0; \
		h = dh; \
	}


void _ScrBlit32to15(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u32 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u16*)(scr + (y+sy) * sp) + sx;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel32(x, y + gsScr->y, bp, dispfb->fbw);
			*pscr++= ((c & 0x0000f8) <<  7) | /* 0x7c00 */
					 ((c & 0x00f800) >>  6) | /* 0x03e0 */
					 ((c & 0xf80000) >> 19);  /* 0x001f */
		}
	}
}

void _ScrBlit16to15(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u16 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u16*)(scr + (y+sy) * sp) + sx;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel16(x, y + gsScr->y, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 10) | /* 0x7c00 */
					 ((c & 0x03e0)      ) | /* 0x03e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void _ScrBlit16Sto15(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u16 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u16*)(scr + (y+sy) * sp) + sx;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel16S(x, y + gsScr->y, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 10) | /* 0x7c00 */
					 ((c & 0x03e0)      ) | /* 0x03e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void ScrBlit15(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	switch (dispfb->psm) {
		case 0x2: _ScrBlit16to15(scr, dw, dh, sp, gsScr, bp); break;
		case 0xA: _ScrBlit16Sto15(scr, dw, dh, sp, gsScr, bp); break;
		default:  _ScrBlit32to15(scr, dw, dh, sp, gsScr, bp); break;
	}
}

////////////////////
// 16bit

void _ScrBlit32to16(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u32 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u16*)(scr + (y+sy) * sp) + sx;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel32(x, y + gsScr->y, bp, dispfb->fbw);
			*pscr++= ((c & 0x0000f8) <<  8) | /* 0xf800 */
					 ((c & 0x00fc00) >>  5) | /* 0x07e0 */
					 ((c & 0xf80000) >> 19);  /* 0x001f */
		}
	}
}

void _ScrBlit16to16(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u16 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u16*)(scr + (y+sy) * sp) + sx;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel16(x, y + gsScr->y, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 11) | /* 0xf800 */
					 ((c & 0x03e0) <<  1) | /* 0x07e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void _ScrBlit16Sto16(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u16 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u16 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u16*)(scr + (y+sy) * sp) + sx;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel16S(x, y + gsScr->y, bp, dispfb->fbw);
			*pscr++= ((c & 0x001f) << 11) | /* 0xf800 */
					 ((c & 0x03e0) <<  1) | /* 0x07e0 */
					 ((c & 0x7c00) >> 10);  /* 0x001f */
		}
	}
}

void ScrBlit16(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	switch (dispfb->psm) {
		case 0x2: _ScrBlit16to16(scr, dw, dh, sp, gsScr, bp); break;
		case 0xA: _ScrBlit16Sto16(scr, dw, dh, sp, gsScr, bp); break;
		default:  _ScrBlit32to16(scr, dw, dh, sp, gsScr, bp); break;
	}
}

////////////////////
// 32bit

void _ScrBlit32to32(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u8 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u32 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u8 *)(scr + (y+sy) * sp) + sx*4;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel32(x, y + gsScr->y, bp, dispfb->fbw);
			pscr[0] = (c >> 16) & 0xff;
			pscr[1] = (c >>  8) & 0xff;
			pscr[2] = c & 0xff;
			pscr+=4;
		}
	}
}

void _ScrBlit16to32(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u8 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u16 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u8 *)(scr + (y+sy) * sp) + sx*4;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel16(x, y + gsScr->y, bp, dispfb->fbw);
			pscr[0] = (c >> 7) & 0xf8;
			pscr[1] = (c >> 2) & 0xf8;
			pscr[2] = (c << 3) & 0xf8;
			pscr+=4;
		}
	}
}

void _ScrBlit16Sto32(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	u8 *pscr;
	int sx, sy;
	int x, y;
	int w, h;
	u16 c;

	_ScrBlitSET();
	for (y=0; y<h; y++) {
		pscr = (u8 *)(scr + (y+sy) * sp) + sx*4;
		for (x=gsScr->x; x<w+gsScr->x; x++) {
			c = readPixel16S(x, y + gsScr->y, bp, dispfb->fbw);
			pscr[0] = (c >> 7) & 0xf8;
			pscr[1] = (c >> 2) & 0xf8;
			pscr[2] = (c << 3) & 0xf8;
			pscr+=4;
		}
	}
}

void ScrBlit32(u8 *scr, int dw, int dh, int sp, Rect *gsScr, u32 bp) {
	switch (dispfb->psm) {
		case 0x2: _ScrBlit16to32(scr, dw, dh, sp, gsScr, bp); break;
		case 0xA: _ScrBlit16Sto32(scr, dw, dh, sp, gsScr, bp); break;
		default:  _ScrBlit32to32(scr, dw, dh, sp, gsScr, bp); break;
	}
}

void ScrBlit(SDL_Surface *dst) {
	Rect gsScr;
	u32 bp;
	int y;

	if (showfullvram) {
		gsScr.x = 0;
		gsScr.y = 0;
		gsScr.w = dispfb->fbw;
		gsScr.h = dispfb->fbh;
		bp = 0;
	} else {
		gsScr.x = display->x + dispfb->dbx;
		gsScr.y = display->y + dispfb->dby;
		gsScr.w = display->w;
		gsScr.h = display->h;
		bp = dispfb->fbp;
	}

	if (conf.fps) y = 16;
	else y = 0;

	_ScrBlit((u8*)dst->pixels + y * dst->pitch, dst->w, dst->h-y, dst->pitch, &gsScr, bp);
}


void SETScrBlit(int bpp) {
	switch (bpp) {
		case 15:
			if (conf.stretch) _ScrBlit = ScrBlit15S;
			else _ScrBlit = ScrBlit15;
			break;
		case 16:
			if (conf.stretch) _ScrBlit = ScrBlit16S;
			else _ScrBlit = ScrBlit16;
			break;
		case 32:
			if (conf.stretch) _ScrBlit = ScrBlit32S;
			else _ScrBlit = ScrBlit32;
			break;
	}
}
