/*  GSsoft
 *  Copyright (C) 2002-2004  GSsoft Team
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>

#include <SDL/SDL.h>
#include <SDL/SDL_syswm.h>
#include <SDL_gfxPrimitives.h>
//#include <SDL_framerate.h>

#include "GS.h"
#include "Draw.h"
#include "Rec.h"
#include "scale2x.h"

SDL_Surface *surf1x;
SDL_Surface *surf2x;
SDL_Surface *surf;
SDL_SysWMinfo info;
//FPSmanager fps;

void DXclearScr() {
	SDL_FillRect(surf, NULL, 0);
}

s32 DXopen() {
#ifdef __LINUX__
	int screen;
	int w, h;
#endif

	if (SDL_Init(SDL_INIT_VIDEO) < 0) return -1;

	surf = SDL_SetVideoMode(cmode->width, cmode->height, 0, conf.fullscreen ? SDL_FULLSCREEN : 0);
	if (surf == NULL) return -1;

	SDL_VERSION(&info.version);
	if (SDL_GetWMInfo(&info) == 0) return -1;

#ifdef __LINUX__
	info.info.x11.lock_func();
	screen = DefaultScreen(info.info.x11.display);
	w = DisplayWidth(info.info.x11.display, screen);
	h = DisplayHeight(info.info.x11.display, screen);
	XMoveWindow(info.info.x11.display, info.info.x11.wmwindow,
				cmode->width  < w ? (w - cmode->width)  / 2 : 0,
				cmode->height < h ? (h - cmode->height) / 2 : 0);
	info.info.x11.unlock_func();
#endif

	SDL_WM_SetCaption(GStitle, NULL);

	SETScrBlit(surf->format->BitsPerPixel);
	DXclearScr();

	surf1x = surf2x = NULL;

//	SDL_initFramerate(&fps);
//	SDL_setFramerate(&fps, 60);

#ifdef __WIN32__
	return (s32)info.window;
#else
	return (s32)info.info.x11.display;
#endif
}

void DXclose() {
	SDL_Quit();
}

static int scalew, scaleh;

void DXcreateSurfaces() {
	if ((scalew == display->w &&
		 scaleh == display->h) &&
		(surf1x != NULL &&
		 surf2x != NULL)) return;

	scalew = display->w;
	scaleh = display->h;

	if (surf1x) SDL_FreeSurface(surf1x); surf1x = NULL;
	if (surf2x) SDL_FreeSurface(surf2x); surf2x = NULL;

	if (surf->flags & SDL_HWSURFACE) {
		surf1x = SDL_CreateRGBSurface(SDL_HWSURFACE, display->w, display->h, 
									  0, 0, 0, 0, 0);
		if (surf1x == NULL) printf("Error allocating video memory surface, trying with normal memory\n");
	}
	if (surf1x == NULL) {
		surf1x = SDL_CreateRGBSurface(SDL_SWSURFACE, display->w, display->h, 
									  surf->format->BitsPerPixel, surf->format->Rmask,
									  surf->format->Gmask, surf->format->Bmask,
									  surf->format->Amask);
		if (surf1x == NULL) conf.filter = 0;
	}

	if (surf->flags & SDL_HWSURFACE) {
		surf2x = SDL_CreateRGBSurface(SDL_HWSURFACE, display->w*2, display->h*2, 
									  0, 0, 0, 0, 0);
		if (surf2x == NULL) printf("Error allocating video memory surface, trying with normal memory\n");
	}
	if (surf2x == NULL) {
		surf2x = SDL_CreateRGBSurface(SDL_SWSURFACE, display->w*2, display->h*2, 
									  surf->format->BitsPerPixel, surf->format->Rmask,
									  surf->format->Gmask, surf->format->Bmask,
									  surf->format->Amask);
		if (surf2x == NULL) conf.filter = 0;
	}
}

void DXblit() {
	if (conf.filter) DXcreateSurfaces();
	if (conf.filter) {
		SDL_Rect dst;

		SDL_LockSurface(surf1x);
		ScrBlit(surf1x);
		SDL_UnlockSurface(surf1x);

		scale2x(surf1x, surf2x);

		dst.x = 0; if (conf.fps) dst.y = 16; else dst.y = 0;
		dst.w = surf->w; dst.h = surf->h - dst.y;
		SDL_SoftStretch(surf2x, NULL, surf, &dst);

		if (conf.record) {
			SDL_LockSurface(surf);
			recFrame(surf->pixels, surf->pitch, surf->format->BitsPerPixel);
			SDL_UnlockSurface(surf);
		}
	} else {
		SDL_LockSurface(surf);

		ScrBlit(surf);
		if (conf.record) recFrame(surf->pixels, surf->pitch, surf->format->BitsPerPixel);

		SDL_UnlockSurface(surf);
	}
}

void DXupdate() {
	SDL_Rect rect;
	static int to;
	static int fpscount, fpsc, fc;
	int ticks;

	fc++;

	if (conf.fps || conf.frameskip) {
		ticks = SDL_GetTicks();
		fpsc = ticks - to;
		to = ticks;
	}

/*	if (conf.frameskip) {
		if ((fpscount % 3) == 0) {
			norender = 1;
		} else if (norender) {
			norender = 0;
		}
	}*/

	if (norender == 0) {
		DXblit();
	}

	rect.x = 0; rect.y = 0;
	rect.w = cmode->width;
	rect.h = 15;
	if (conf.fps) {
		char title[256];
		char tmp[256];

		if (fpspos < 0) fpspos = 0;
#ifndef GS_LOG
		if (fpspos > 6) fpspos = 6;
#else
		if (fpspos > 7) fpspos = 7;
#endif
		switch (fpspos) {
			case 0: // FrameSkip
				sprintf(tmp, "Frameskip %s", conf.frameskip == 1 ? "On" : "Off");
				norender = 0;
				break;
			case 1: // Stretch
				sprintf(tmp, "Stretch %s", conf.stretch == 1 ? "On" : "Off");
				break;
			case 2: // Misc stuff
				sprintf(tmp, "GSmode %dx%d", display->w, display->h);
				break;
			case 3: // ShowFullVRam
				sprintf(tmp, "ShowFullVRam %s", showfullvram == 1 ? "On" : "Off");
				break;
			case 4: // NoAlphaBlending
				sprintf(tmp, "NoAlphaBlending %s", noabe == 1 ? "On" : "Off");
				break;
			case 5: // DisableRendering
				sprintf(tmp, "DisableRendering %s", norender == 1 ? "On" : "Off");
				break;
			case 6: // WireFrame
				sprintf(tmp, "WireFrame %s", wireframe == 1 ? "On" : "Off");
				break;
#ifdef GS_LOG
			case 7: // Log
				sprintf(tmp, "Log %s", conf.log == 1 ? "On" : "Off");
				break;
#endif
		}
		if (fpspress) {
			switch (fpspos) {
				case 0:
					conf.frameskip = 1 - conf.frameskip;
					norender = 0;
					break;
				case 1:
					conf.stretch = 1 - conf.stretch;
					SETScrBlit(surf->format->BitsPerPixel);
					DXclearScr();
					fpspress = 0;
					return;
				case 3:
					DXclearScr();
					showfullvram = 1 - showfullvram;
					break;
				case 4:
					noabe = 1 - noabe;
					break;
				case 5:
					norender = 1 - norender;
					break;
				case 6:
					wireframe = 1 - wireframe;
					break;
#ifdef GS_LOG
				case 7:
					conf.log = 1 - conf.log;
					break;
#endif
			}
			fpspress = 0;
		}

#ifdef __x86_64__
		sprintf(title,"FPS %d -- %s - FC: %d", 1000 / fpsc, tmp, fc);
#else
		sprintf(title,"FPS %.2f -- %s - FC: %d", 1000.0 / fpsc, tmp, fc);
#endif

		SDL_FillRect(surf, &rect, 0);
		stringRGBA(surf, 4, rect.y+4, title, 0, 0xff, 0, 0xff);
		rect.y+= 15;
	}
/*	if (gsmsg[0]) {
		SDL_FillRect(surf, &rect, 0);
		stringRGBA(surf, 4, rect.y+4, gsmsg, 0, 0xff, 0, 0xff);
		if (SDL_GetTicks() >= gsmsgt) gsmsg[0] = 0;
	}*/

	SDL_UpdateRect(surf, 0, 0, 0, 0);
#ifdef __WIN32__
	SDL_PumpEvents();
#endif
}

int DXgetModes() {
	SDL_Rect **rects;
	int i;

	if (SDL_Init(SDL_INIT_VIDEO) < 0) return -1;

	memset(modes, 0, sizeof(modes));
	rects = SDL_ListModes(NULL, SDL_FULLSCREEN);
	if (rects == NULL) return -1;

	for (i=0; i<64; i++) {
		if (rects[i] == NULL) break;
		modes[i].width  = rects[i]->w;
		modes[i].height = rects[i]->h;
	}

	SDL_Quit();

	return i;
}

void CALLBACK GSprintf(int timeout, char *fmt, ...) {
/*	va_list list;

	va_start(list,fmt);
	vsprintf(gsmsg,fmt,list);
	va_end(list);
	gsmsgt = SDL_GetTicks()+timeout*1000;*/
}

void CALLBACK GSgetDriverInfo(GSdriverInfo *info) {
	strcpy(info->name, "sdl");
	info->common = surf;
}
